@@ -410,6 +410,7 @@ static int translateKey(unsigned int key, bool apply_keymap)
                 K('[', LEFT_BRACKET);
                 K('\\', BACKSLASH);
                 K(']', RIGHT_BRACKET);
+                K('^', CIRCUMFLEX);
                 K('_', UNDERSCORE);
                 K('`', GRAVE_ACCENT);
                 K(PARAGRAPH_UTF_8, PARAGRAPH);
@@ -658,6 +659,7 @@ static const NSRange kEmptyRange = { NSNotFound, 0 };
     NSTrackingArea* trackingArea;
     NSMutableAttributedString* markedText;
     NSRect markedRect;
+    NSString *input_source_at_last_key_event;
 }
 
 - (void) removeGLFWWindow;
@@ -676,6 +678,7 @@ static const NSRange kEmptyRange = { NSNotFound, 0 };
         trackingArea = nil;
         markedText = [[NSMutableAttributedString alloc] init];
         markedRect = NSMakeRect(0.0, 0.0, 0.0, 0.0);
+        input_source_at_last_key_event = nil;
 
         [self updateTrackingAreas];
         [self registerForDraggedTypes:@[NSPasteboardTypeFileURL, NSPasteboardTypeString]];
@@ -688,6 +691,7 @@ static const NSRange kEmptyRange = { NSNotFound, 0 };
 {
     [trackingArea release];
     [markedText release];
+    if (input_source_at_last_key_event) [input_source_at_last_key_event release];
     [super dealloc];
 }
 
@@ -964,12 +968,19 @@ is_ascii_control_char(char x) {
 
 - (void)keyDown:(NSEvent *)event
 {
+    const bool previous_has_marked_text = [self hasMarkedText];
+    bool input_source_changed = false;
+    NSTextInputContext *inpctx = [NSTextInputContext currentInputContext];
+    if (inpctx && (!input_source_at_last_key_event || ![input_source_at_last_key_event isEqualToString:inpctx.selectedKeyboardInputSource])) {
+        input_source_at_last_key_event = [inpctx.selectedKeyboardInputSource retain];
+        input_source_changed = true;
+    }
+
     const unsigned int keycode = [event keyCode];
     const NSUInteger flags = [event modifierFlags];
     const int mods = translateFlags(flags);
     const int key = translateKey(keycode, true);
     const bool process_text = !window->ns.textInputFilterCallback || window->ns.textInputFilterCallback(key, mods, keycode, flags) != 1;
-    const bool previous_has_marked_text = [self hasMarkedText];
     [self unmarkText];
     _glfw.ns.text[0] = 0;
     GLFWkeyevent glfw_keyevent;
@@ -983,6 +994,14 @@ is_ascii_control_char(char x) {
             [self interpretKeyEvents:[NSArray arrayWithObject:event]];
         }
     } else {
+        if (input_source_changed) {
+            debug_key(@"Input source changed, clearing pre-edit text and resetting deadkey state\n");
+            glfw_keyevent.text = NULL;
+            glfw_keyevent.ime_state = 1;
+            window->ns.deadKeyState = 0;
+            _glfwInputKeyboard(window, &glfw_keyevent); // clear pre-edit text
+        }
+
         static UniChar text[256];
         UniCharCount char_count = 0;
         const bool in_compose_sequence = window->ns.deadKeyState != 0;
@@ -1114,7 +1133,7 @@ is_ascii_control_char(char x) {
             break;
     }
 
-    _glfwInputScroll(window, deltaX, deltaY, flags);
+    _glfwInputScroll(window, deltaX, deltaY, flags, translateFlags([event modifierFlags]));
 }
 
 - (NSDragOperation)draggingEntered:(id <NSDraggingInfo>)sender
@@ -1140,27 +1159,20 @@ is_ascii_control_char(char x) {
     const NSUInteger count = [objs count];
     if (count)
     {
-        char** paths = calloc(count, sizeof(char*));
-
         for (NSUInteger i = 0;  i < count;  i++)
         {
             id obj = objs[i];
             if ([obj isKindOfClass:[NSURL class]]) {
-                paths[i] = _glfw_strdup([obj fileSystemRepresentation]);
+                const char *path = [obj fileSystemRepresentation];
+                _glfwInputDrop(window, "text/plain;charset=utf-8", path, strlen(path));
             } else if ([obj isKindOfClass:[NSString class]]) {
-                paths[i] = _glfw_strdup([obj UTF8String]);
+                const char *text = [obj UTF8String];
+                _glfwInputDrop(window, "text/plain;charset=utf-8", text, strlen(text));
             } else {
                 _glfwInputError(GLFW_PLATFORM_ERROR,
                                 "Cocoa: Object is neither a URL nor a string");
-                paths[i] = _glfw_strdup("");
             }
         }
-
-        _glfwInputDrop(window, (int) count, (const char**) paths);
-
-        for (NSUInteger i = 0;  i < count;  i++)
-            free(paths[i]);
-        free(paths);
     }
 
     return YES;
@@ -1448,6 +1460,11 @@ int _glfwPlatformCreateWindow(_GLFWwindow* window,
         }
         else if (ctxconfig->source == GLFW_EGL_CONTEXT_API)
         {
+            // EGL implementation on macOS use CALayer* EGLNativeWindowType so we
+            // need to get the layer for EGL window surface creation.
+            [window->ns.view setWantsLayer:YES];
+            window->ns.layer = [window->ns.view layer];
+
             if (!_glfwInitEGL())
                 return false;
             if (!_glfwCreateContextEGL(window, ctxconfig, fbconfig))
@@ -1498,7 +1515,7 @@ void _glfwPlatformDestroyWindow(_GLFWwindow* window)
     window->ns.object = nil;
 }
 
-void _glfwPlatformSetWindowTitle(_GLFWwindow* window UNUSED, const char *title)
+void _glfwPlatformSetWindowTitle(_GLFWwindow* window UNUSED, const char* title)
 {
     NSString* string = @(title);
     [window->ns.object setTitle:string];
@@ -1510,7 +1527,8 @@ void _glfwPlatformSetWindowTitle(_GLFWwindow* window UNUSED, const char *title)
 void _glfwPlatformSetWindowIcon(_GLFWwindow* window UNUSED,
                                 int count UNUSED, const GLFWimage* images UNUSED)
 {
-    // Regular windows do not have icons
+    _glfwInputError(GLFW_FEATURE_UNAVAILABLE,
+                    "Cocoa: Regular windows do not have icons on macOS");
 }
 
 void _glfwPlatformGetWindowPos(_GLFWwindow* window, int* xpos, int* ypos)
@@ -1577,10 +1595,18 @@ void _glfwPlatformSetWindowSizeLimits(_GLFWwindow* window,
 
 void _glfwPlatformSetWindowAspectRatio(_GLFWwindow* window, int numer, int denom)
 {
-    if (numer == GLFW_DONT_CARE || denom == GLFW_DONT_CARE)
-        [window->ns.object setResizeIncrements:NSMakeSize(1.0, 1.0)];
-    else
+    if (numer != GLFW_DONT_CARE && denom != GLFW_DONT_CARE)
         [window->ns.object setContentAspectRatio:NSMakeSize(numer, denom)];
+    else
+        [window->ns.object setResizeIncrements:NSMakeSize(1.0, 1.0)];
+}
+
+void _glfwPlatformSetWindowSizeIncrements(_GLFWwindow* window, int widthincr, int heightincr)
+{
+    if (widthincr != GLFW_DONT_CARE && heightincr != GLFW_DONT_CARE)
+        [window->ns.object setResizeIncrements:NSMakeSize(widthincr, heightincr)];
+    else
+        [window->ns.object setResizeIncrements:NSMakeSize(1.0, 1.0)];
 }
 
 void _glfwPlatformGetFramebufferSize(_GLFWwindow* window, int* width, int* height)
@@ -1828,6 +1854,11 @@ void _glfwPlatformSetWindowFloating(_GLFWwindow* window, bool enabled)
         [window->ns.object setLevel:NSNormalWindowLevel];
 }
 
+void _glfwPlatformSetWindowMousePassthrough(_GLFWwindow* window, bool enabled)
+{
+    [window->ns.object setIgnoresMouseEvents:enabled];
+}
+
 float _glfwPlatformGetWindowOpacity(_GLFWwindow* window)
 {
     return (float) [window->ns.object alphaValue];
@@ -1838,6 +1869,17 @@ void _glfwPlatformSetWindowOpacity(_GLFWwindow* window, float opacity)
     [window->ns.object setAlphaValue:opacity];
 }
 
+void _glfwPlatformSetRawMouseMotion(_GLFWwindow *window UNUSED, bool enabled UNUSED)
+{
+    _glfwInputError(GLFW_FEATURE_UNIMPLEMENTED,
+                    "Cocoa: Raw mouse motion not yet implemented");
+}
+
+bool _glfwPlatformRawMouseMotionSupported(void)
+{
+    return false;
+}
+
 void
 _glfwDispatchRenderFrame(CGDirectDisplayID displayID) {
     _GLFWwindow *w = _glfw.windowListHead;
@@ -2068,13 +2110,59 @@ const char* _glfwPlatformGetClipboardString(void)
     return _glfw.ns.clipboardString;
 }
 
+EGLenum _glfwPlatformGetEGLPlatform(EGLint** attribs)
+{
+    if (_glfw.egl.ANGLE_platform_angle)
+    {
+        int type = 0;
+
+        if (_glfw.egl.ANGLE_platform_angle_opengl)
+        {
+            if (_glfw.hints.init.angleType == GLFW_ANGLE_PLATFORM_TYPE_OPENGL)
+                type = EGL_PLATFORM_ANGLE_TYPE_OPENGL_ANGLE;
+        }
+
+        if (_glfw.egl.ANGLE_platform_angle_metal)
+        {
+            if (_glfw.hints.init.angleType == GLFW_ANGLE_PLATFORM_TYPE_METAL)
+                type = EGL_PLATFORM_ANGLE_TYPE_METAL_ANGLE;
+        }
+
+        if (type)
+        {
+            *attribs = calloc(3, sizeof(EGLint));
+            (*attribs)[0] = EGL_PLATFORM_ANGLE_TYPE_ANGLE;
+            (*attribs)[1] = type;
+            (*attribs)[2] = EGL_NONE;
+            return EGL_PLATFORM_ANGLE_ANGLE;
+        }
+    }
+
+    return 0;
+}
+
+EGLNativeDisplayType _glfwPlatformGetEGLNativeDisplay(void)
+{
+    return EGL_DEFAULT_DISPLAY;
+}
+
+EGLNativeWindowType _glfwPlatformGetEGLNativeWindow(_GLFWwindow* window)
+{
+    return window->ns.layer;
+}
+
 void _glfwPlatformGetRequiredInstanceExtensions(char** extensions)
 {
-    if (!_glfw.vk.KHR_surface || !_glfw.vk.MVK_macos_surface)
-        return;
-
-    extensions[0] = "VK_KHR_surface";
-    extensions[1] = "VK_MVK_macos_surface";
+    if (_glfw.vk.KHR_surface && _glfw.vk.EXT_metal_surface)
+    {
+        extensions[0] = "VK_KHR_surface";
+        extensions[1] = "VK_EXT_metal_surface";
+    }
+    else if (_glfw.vk.KHR_surface && _glfw.vk.MVK_macos_surface)
+    {
+        extensions[0] = "VK_KHR_surface";
+        extensions[1] = "VK_MVK_macos_surface";
+    }
 }
 
 int _glfwPlatformGetPhysicalDevicePresentationSupport(VkInstance instance UNUSED,
@@ -2090,19 +2178,6 @@ VkResult _glfwPlatformCreateWindowSurface(VkInstance instance,
                                           VkSurfaceKHR* surface)
 {
 #if MAC_OS_X_VERSION_MAX_ALLOWED >= 101100
-    VkResult err;
-    VkMacOSSurfaceCreateInfoMVK sci;
-    PFN_vkCreateMacOSSurfaceMVK vkCreateMacOSSurfaceMVK;
-
-    vkCreateMacOSSurfaceMVK = (PFN_vkCreateMacOSSurfaceMVK)
-        vkGetInstanceProcAddr(instance, "vkCreateMacOSSurfaceMVK");
-    if (!vkCreateMacOSSurfaceMVK)
-    {
-        _glfwInputError(GLFW_API_UNAVAILABLE,
-                        "Cocoa: Vulkan instance missing VK_MVK_macos_surface extension");
-        return VK_ERROR_EXTENSION_NOT_PRESENT;
-    }
-
     // HACK: Dynamically load Core Animation to avoid adding an extra
     //       dependency for the majority who don't use MoltenVK
     NSBundle* bundle = [NSBundle bundleWithPath:@"/System/Library/Frameworks/QuartzCore.framework"];
@@ -2128,11 +2203,49 @@ VkResult _glfwPlatformCreateWindowSurface(VkInstance instance,
     [window->ns.view setLayer:window->ns.layer];
     [window->ns.view setWantsLayer:YES];
 
-    memset(&sci, 0, sizeof(sci));
-    sci.sType = VK_STRUCTURE_TYPE_MACOS_SURFACE_CREATE_INFO_MVK;
-    sci.pView = window->ns.view;
+    VkResult err;
+
+    if (_glfw.vk.EXT_metal_surface)
+    {
+        VkMetalSurfaceCreateInfoEXT sci;
+
+        PFN_vkCreateMetalSurfaceEXT vkCreateMetalSurfaceEXT;
+        vkCreateMetalSurfaceEXT = (PFN_vkCreateMetalSurfaceEXT)
+            vkGetInstanceProcAddr(instance, "vkCreateMetalSurfaceEXT");
+        if (!vkCreateMetalSurfaceEXT)
+        {
+            _glfwInputError(GLFW_API_UNAVAILABLE,
+                            "Cocoa: Vulkan instance missing VK_EXT_metal_surface extension");
+            return VK_ERROR_EXTENSION_NOT_PRESENT;
+        }
+
+        memset(&sci, 0, sizeof(sci));
+        sci.sType = VK_STRUCTURE_TYPE_METAL_SURFACE_CREATE_INFO_EXT;
+        sci.pLayer = window->ns.layer;
+
+        err = vkCreateMetalSurfaceEXT(instance, &sci, allocator, surface);
+    }
+    else
+    {
+        VkMacOSSurfaceCreateInfoMVK sci;
+
+        PFN_vkCreateMacOSSurfaceMVK vkCreateMacOSSurfaceMVK;
+        vkCreateMacOSSurfaceMVK = (PFN_vkCreateMacOSSurfaceMVK)
+            vkGetInstanceProcAddr(instance, "vkCreateMacOSSurfaceMVK");
+        if (!vkCreateMacOSSurfaceMVK)
+        {
+            _glfwInputError(GLFW_API_UNAVAILABLE,
+                            "Cocoa: Vulkan instance missing VK_MVK_macos_surface extension");
+            return VK_ERROR_EXTENSION_NOT_PRESENT;
+        }
+
+        memset(&sci, 0, sizeof(sci));
+        sci.sType = VK_STRUCTURE_TYPE_MACOS_SURFACE_CREATE_INFO_MVK;
+        sci.pView = window->ns.view;
+
+        err = vkCreateMacOSSurfaceMVK(instance, &sci, allocator, surface);
+    }
 
-    err = vkCreateMacOSSurfaceMVK(instance, &sci, allocator, surface);
     if (err)
     {
         _glfwInputError(GLFW_PLATFORM_ERROR,
@@ -2178,9 +2291,9 @@ GLFWAPI void glfwCocoaRequestRenderFrame(GLFWwindow *w, GLFWcocoarenderframefun
     requestRenderFrame((_GLFWwindow*)w, callback);
 }
 
-GLFWAPI void glfwGetCocoaKeyEquivalent(int glfw_key, int glfw_mods, unsigned short *cocoa_key, int *cocoa_mods) {
-    *cocoa_key = 0;
+GLFWAPI void glfwGetCocoaKeyEquivalent(int glfw_key, int glfw_mods, char *cocoa_key, size_t key_sz, int *cocoa_mods) {
     *cocoa_mods = 0;
+    memset(cocoa_key, 0, key_sz);
 
     if (glfw_mods & GLFW_MOD_SHIFT)
         *cocoa_mods |= NSEventModifierFlagShift;
@@ -2193,147 +2306,156 @@ GLFWAPI void glfwGetCocoaKeyEquivalent(int glfw_key, int glfw_mods, unsigned sho
     if (glfw_mods & GLFW_MOD_CAPS_LOCK)
         *cocoa_mods |= NSEventModifierFlagCapsLock;
 
+    uint32_t utf_8_key = 0;
+    unichar utf_16_key = 0;
+
 START_ALLOW_CASE_RANGE
     switch(glfw_key) {
-#define K(ch, name) case GLFW_KEY_##name: *cocoa_key = ch; break;
-        K('!', EXCLAM);
-        K('"', DOUBLE_QUOTE);
-        K('#', NUMBER_SIGN);
-        K('$', DOLLAR);
-        K('&', AMPERSAND);
-        K('\'', APOSTROPHE);
-        K('(', PARENTHESIS_LEFT);
-        K(')', PARENTHESIS_RIGHT);
-        K('+', PLUS);
-        K(',', COMMA);
-        K('-', MINUS);
-        K('.', PERIOD);
-        K('/', SLASH);
-        K('0', 0);
-        K('1', 1);
-        K('2', 2);
-        K('3', 3);
-        K('5', 5);
-        K('6', 6);
-        K('7', 7);
-        K('8', 8);
-        K('9', 9);
-        K(':', COLON);
-        K(';', SEMICOLON);
-        K('<', LESS);
-        K('=', EQUAL);
-        K('>', GREATER);
-        K('@', AT);
-        K('[', LEFT_BRACKET);
-        K('\\', BACKSLASH);
-        K(']', RIGHT_BRACKET);
-        K('_', UNDERSCORE);
-        K('`', GRAVE_ACCENT);
-        K('a', A);
-        K('b', B);
-        K('c', C);
-        K('d', D);
-        K('e', E);
-        K('f', F);
-        K('g', G);
-        K('h', H);
-        K('i', I);
-        K('j', J);
-        K('k', K);
-        K('l', L);
-        K('m', M);
-        K('n', N);
-        K('o', O);
-        K('p', P);
-        K('q', Q);
-        K('r', R);
-        K('s', S);
-        K('t', T);
-        K('u', U);
-        K('v', V);
-        K('w', W);
-        K('x', X);
-        K('y', Y);
-        K('z', Z);
-        K(PARAGRAPH_UTF_8, PARAGRAPH);
-        K(MASCULINE_UTF_8, MASCULINE);
-        K(S_SHARP_UTF_8, S_SHARP);
-        K(A_GRAVE_LOWER_CASE_UTF_8, A_GRAVE);
-        K(A_DIAERESIS_LOWER_CASE_UTF_8, A_DIAERESIS);
-        K(A_RING_LOWER_CASE_UTF_8, A_RING);
-        K(AE_LOWER_CASE_UTF_8, AE);
-        K(C_CEDILLA_LOWER_CASE_UTF_8, C_CEDILLA);
-        K(E_GRAVE_LOWER_CASE_UTF_8, E_GRAVE);
-        K(E_ACUTE_LOWER_CASE_UTF_8, E_ACUTE);
-        K(I_GRAVE_LOWER_CASE_UTF_8, I_GRAVE);
-        K(N_TILDE_LOWER_CASE_UTF_8, N_TILDE);
-        K(O_GRAVE_LOWER_CASE_UTF_8, O_GRAVE);
-        K(O_DIAERESIS_LOWER_CASE_UTF_8, O_DIAERESIS);
-        K(O_SLASH_LOWER_CASE_UTF_8, O_SLASH);
-        K(U_GRAVE_LOWER_CASE_UTF_8, U_GRAVE);
-        K(U_DIAERESIS_LOWER_CASE_UTF_8, U_DIAERESIS);
-        K(CYRILLIC_A_LOWER_CASE_UTF_8, CYRILLIC_A);
-        K(CYRILLIC_BE_LOWER_CASE_UTF_8, CYRILLIC_BE);
-        K(CYRILLIC_VE_LOWER_CASE_UTF_8, CYRILLIC_VE);
-        K(CYRILLIC_GHE_LOWER_CASE_UTF_8, CYRILLIC_GHE);
-        K(CYRILLIC_DE_LOWER_CASE_UTF_8, CYRILLIC_DE);
-        K(CYRILLIC_IE_LOWER_CASE_UTF_8, CYRILLIC_IE);
-        K(CYRILLIC_ZHE_LOWER_CASE_UTF_8, CYRILLIC_ZHE);
-        K(CYRILLIC_ZE_LOWER_CASE_UTF_8, CYRILLIC_ZE);
-        K(CYRILLIC_I_LOWER_CASE_UTF_8, CYRILLIC_I);
-        K(CYRILLIC_SHORT_I_LOWER_CASE_UTF_8, CYRILLIC_SHORT_I);
-        K(CYRILLIC_KA_LOWER_CASE_UTF_8, CYRILLIC_KA);
-        K(CYRILLIC_EL_LOWER_CASE_UTF_8, CYRILLIC_EL);
-        K(CYRILLIC_EM_LOWER_CASE_UTF_8, CYRILLIC_EM);
-        K(CYRILLIC_EN_LOWER_CASE_UTF_8, CYRILLIC_EN);
-        K(CYRILLIC_O_LOWER_CASE_UTF_8, CYRILLIC_O);
-        K(CYRILLIC_PE_LOWER_CASE_UTF_8, CYRILLIC_PE);
-        K(CYRILLIC_ER_LOWER_CASE_UTF_8, CYRILLIC_ER);
-        K(CYRILLIC_ES_LOWER_CASE_UTF_8, CYRILLIC_ES);
-        K(CYRILLIC_TE_LOWER_CASE_UTF_8, CYRILLIC_TE);
-        K(CYRILLIC_U_LOWER_CASE_UTF_8, CYRILLIC_U);
-        K(CYRILLIC_EF_LOWER_CASE_UTF_8, CYRILLIC_EF);
-        K(CYRILLIC_HA_LOWER_CASE_UTF_8, CYRILLIC_HA);
-        K(CYRILLIC_TSE_LOWER_CASE_UTF_8, CYRILLIC_TSE);
-        K(CYRILLIC_CHE_LOWER_CASE_UTF_8, CYRILLIC_CHE);
-        K(CYRILLIC_SHA_LOWER_CASE_UTF_8, CYRILLIC_SHA);
-        K(CYRILLIC_SHCHA_LOWER_CASE_UTF_8, CYRILLIC_SHCHA);
-        K(CYRILLIC_HARD_SIGN_LOWER_CASE_UTF_8, CYRILLIC_HARD_SIGN);
-        K(CYRILLIC_YERU_LOWER_CASE_UTF_8, CYRILLIC_YERU);
-        K(CYRILLIC_SOFT_SIGN_LOWER_CASE_UTF_8, CYRILLIC_SOFT_SIGN);
-        K(CYRILLIC_E_LOWER_CASE_UTF_8, CYRILLIC_E);
-        K(CYRILLIC_YU_LOWER_CASE_UTF_8, CYRILLIC_YU);
-        K(CYRILLIC_YA_LOWER_CASE_UTF_8, CYRILLIC_YA);
-        K(CYRILLIC_IO_LOWER_CASE_UTF_8, CYRILLIC_IO);
+#define K8(ch, name) case GLFW_KEY_##name: utf_8_key = ch; break;
+#define K16(ch, name) case GLFW_KEY_##name: utf_16_key = ch; break;
+        K8('!', EXCLAM);
+        K8('"', DOUBLE_QUOTE);
+        K8('#', NUMBER_SIGN);
+        K8('$', DOLLAR);
+        K8('&', AMPERSAND);
+        K8('\'', APOSTROPHE);
+        K8('(', PARENTHESIS_LEFT);
+        K8(')', PARENTHESIS_RIGHT);
+        K8('+', PLUS);
+        K8(',', COMMA);
+        K8('-', MINUS);
+        K8('.', PERIOD);
+        K8('/', SLASH);
+        K8('0', 0);
+        K8('1', 1);
+        K8('2', 2);
+        K8('3', 3);
+        K8('5', 5);
+        K8('6', 6);
+        K8('7', 7);
+        K8('8', 8);
+        K8('9', 9);
+        K8(':', COLON);
+        K8(';', SEMICOLON);
+        K8('<', LESS);
+        K8('=', EQUAL);
+        K8('>', GREATER);
+        K8('@', AT);
+        K8('[', LEFT_BRACKET);
+        K8('\\', BACKSLASH);
+        K8(']', RIGHT_BRACKET);
+        K8('^', CIRCUMFLEX);
+        K8('_', UNDERSCORE);
+        K8('`', GRAVE_ACCENT);
+        K8('a', A);
+        K8('b', B);
+        K8('c', C);
+        K8('d', D);
+        K8('e', E);
+        K8('f', F);
+        K8('g', G);
+        K8('h', H);
+        K8('i', I);
+        K8('j', J);
+        K8('k', K);
+        K8('l', L);
+        K8('m', M);
+        K8('n', N);
+        K8('o', O);
+        K8('p', P);
+        K8('q', Q);
+        K8('r', R);
+        K8('s', S);
+        K8('t', T);
+        K8('u', U);
+        K8('v', V);
+        K8('w', W);
+        K8('x', X);
+        K8('y', Y);
+        K8('z', Z);
+        K8(PARAGRAPH_UTF_8, PARAGRAPH);
+        K8(MASCULINE_UTF_8, MASCULINE);
+        K8(S_SHARP_UTF_8, S_SHARP);
+        K8(A_GRAVE_LOWER_CASE_UTF_8, A_GRAVE);
+        K8(A_DIAERESIS_LOWER_CASE_UTF_8, A_DIAERESIS);
+        K8(A_RING_LOWER_CASE_UTF_8, A_RING);
+        K8(AE_LOWER_CASE_UTF_8, AE);
+        K8(C_CEDILLA_LOWER_CASE_UTF_8, C_CEDILLA);
+        K8(E_GRAVE_LOWER_CASE_UTF_8, E_GRAVE);
+        K8(E_ACUTE_LOWER_CASE_UTF_8, E_ACUTE);
+        K8(I_GRAVE_LOWER_CASE_UTF_8, I_GRAVE);
+        K8(N_TILDE_LOWER_CASE_UTF_8, N_TILDE);
+        K8(O_GRAVE_LOWER_CASE_UTF_8, O_GRAVE);
+        K8(O_DIAERESIS_LOWER_CASE_UTF_8, O_DIAERESIS);
+        K8(O_SLASH_LOWER_CASE_UTF_8, O_SLASH);
+        K8(U_GRAVE_LOWER_CASE_UTF_8, U_GRAVE);
+        K8(U_DIAERESIS_LOWER_CASE_UTF_8, U_DIAERESIS);
+        K8(CYRILLIC_A_LOWER_CASE_UTF_8, CYRILLIC_A);
+        K8(CYRILLIC_BE_LOWER_CASE_UTF_8, CYRILLIC_BE);
+        K8(CYRILLIC_VE_LOWER_CASE_UTF_8, CYRILLIC_VE);
+        K8(CYRILLIC_GHE_LOWER_CASE_UTF_8, CYRILLIC_GHE);
+        K8(CYRILLIC_DE_LOWER_CASE_UTF_8, CYRILLIC_DE);
+        K8(CYRILLIC_IE_LOWER_CASE_UTF_8, CYRILLIC_IE);
+        K8(CYRILLIC_ZHE_LOWER_CASE_UTF_8, CYRILLIC_ZHE);
+        K8(CYRILLIC_ZE_LOWER_CASE_UTF_8, CYRILLIC_ZE);
+        K8(CYRILLIC_I_LOWER_CASE_UTF_8, CYRILLIC_I);
+        K8(CYRILLIC_SHORT_I_LOWER_CASE_UTF_8, CYRILLIC_SHORT_I);
+        K8(CYRILLIC_KA_LOWER_CASE_UTF_8, CYRILLIC_KA);
+        K8(CYRILLIC_EL_LOWER_CASE_UTF_8, CYRILLIC_EL);
+        K8(CYRILLIC_EM_LOWER_CASE_UTF_8, CYRILLIC_EM);
+        K8(CYRILLIC_EN_LOWER_CASE_UTF_8, CYRILLIC_EN);
+        K8(CYRILLIC_O_LOWER_CASE_UTF_8, CYRILLIC_O);
+        K8(CYRILLIC_PE_LOWER_CASE_UTF_8, CYRILLIC_PE);
+        K8(CYRILLIC_ER_LOWER_CASE_UTF_8, CYRILLIC_ER);
+        K8(CYRILLIC_ES_LOWER_CASE_UTF_8, CYRILLIC_ES);
+        K8(CYRILLIC_TE_LOWER_CASE_UTF_8, CYRILLIC_TE);
+        K8(CYRILLIC_U_LOWER_CASE_UTF_8, CYRILLIC_U);
+        K8(CYRILLIC_EF_LOWER_CASE_UTF_8, CYRILLIC_EF);
+        K8(CYRILLIC_HA_LOWER_CASE_UTF_8, CYRILLIC_HA);
+        K8(CYRILLIC_TSE_LOWER_CASE_UTF_8, CYRILLIC_TSE);
+        K8(CYRILLIC_CHE_LOWER_CASE_UTF_8, CYRILLIC_CHE);
+        K8(CYRILLIC_SHA_LOWER_CASE_UTF_8, CYRILLIC_SHA);
+        K8(CYRILLIC_SHCHA_LOWER_CASE_UTF_8, CYRILLIC_SHCHA);
+        K8(CYRILLIC_HARD_SIGN_LOWER_CASE_UTF_8, CYRILLIC_HARD_SIGN);
+        K8(CYRILLIC_YERU_LOWER_CASE_UTF_8, CYRILLIC_YERU);
+        K8(CYRILLIC_SOFT_SIGN_LOWER_CASE_UTF_8, CYRILLIC_SOFT_SIGN);
+        K8(CYRILLIC_E_LOWER_CASE_UTF_8, CYRILLIC_E);
+        K8(CYRILLIC_YU_LOWER_CASE_UTF_8, CYRILLIC_YU);
+        K8(CYRILLIC_YA_LOWER_CASE_UTF_8, CYRILLIC_YA);
+        K8(CYRILLIC_IO_LOWER_CASE_UTF_8, CYRILLIC_IO);
 
-        K(0x35, ESCAPE);
-        K('\r', ENTER);
-        K('\t', TAB);
-        K(NSBackspaceCharacter, BACKSPACE);
-        K(NSInsertFunctionKey, INSERT);
-        K(NSDeleteCharacter, DELETE);
-        K(NSLeftArrowFunctionKey, LEFT);
-        K(NSRightArrowFunctionKey, RIGHT);
-        K(NSUpArrowFunctionKey, UP);
-        K(NSDownArrowFunctionKey, DOWN);
-        K(NSPageUpFunctionKey, PAGE_UP);
-        K(NSPageDownFunctionKey, PAGE_DOWN);
-        K(NSHomeFunctionKey, HOME);
-        K(NSEndFunctionKey, END);
-        K(NSPrintFunctionKey, PRINT_SCREEN);
+        K8(0x35, ESCAPE);
+        K8('\r', ENTER);
+        K8('\t', TAB);
+        K16(NSBackspaceCharacter, BACKSPACE);
+        K16(NSInsertFunctionKey, INSERT);
+        K16(NSDeleteCharacter, DELETE);
+        K16(NSLeftArrowFunctionKey, LEFT);
+        K16(NSRightArrowFunctionKey, RIGHT);
+        K16(NSUpArrowFunctionKey, UP);
+        K16(NSDownArrowFunctionKey, DOWN);
+        K16(NSPageUpFunctionKey, PAGE_UP);
+        K16(NSPageDownFunctionKey, PAGE_DOWN);
+        K16(NSHomeFunctionKey, HOME);
+        K16(NSEndFunctionKey, END);
+        K16(NSPrintFunctionKey, PRINT_SCREEN);
         case GLFW_KEY_F1 ... GLFW_KEY_F24:
-            *cocoa_key = NSF1FunctionKey + (glfw_key - GLFW_KEY_F1); break;
-        case GLFW_KEY_KP_0 ... GLFW_KEY_KP_9:
-            *cocoa_key = NSEventModifierFlagNumericPad | (0x52 + (glfw_key - GLFW_KEY_KP_0)); break;
-        K((unichar)(0x41|NSEventModifierFlagNumericPad), KP_DECIMAL);
-        K((unichar)(0x43|NSEventModifierFlagNumericPad), KP_MULTIPLY);
-        K((unichar)(0x45|NSEventModifierFlagNumericPad), KP_ADD);
-        K((unichar)(0x4B|NSEventModifierFlagNumericPad), KP_DIVIDE);
-        K((unichar)(0x4E|NSEventModifierFlagNumericPad), KP_SUBTRACT);
-        K((unichar)(0x51|NSEventModifierFlagNumericPad), KP_EQUAL);
-#undef K
+            utf_16_key = NSF1FunctionKey + (glfw_key - GLFW_KEY_F1); break;
+#undef K8
+#undef K16
 END_ALLOW_CASE_RANGE
     }
+    if (utf_16_key != 0) {
+         strncpy(cocoa_key, [[NSString stringWithCharacters:&utf_16_key length:1] UTF8String], key_sz - 1);
+    } else {
+        unsigned str_pos = 0;
+        for (unsigned i = 0; i < 4 && str_pos < key_sz - 1; i++) {
+            uint8_t byte = (utf_8_key >> 24) & 0xff;
+            utf_8_key <<= 8;
+            if (byte != 0) cocoa_key[str_pos++] = byte;
+        }
+        cocoa_key[str_pos] = 0;
+    }
 }
 
 
