@@ -55,10 +55,11 @@ typedef enum MouseTrackingModes { NO_TRACKING, BUTTON_MODE, MOTION_MODE, ANY_MOD
 typedef enum MouseTrackingProtocols { NORMAL_PROTOCOL, UTF8_PROTOCOL, SGR_PROTOCOL, URXVT_PROTOCOL} MouseTrackingProtocol;
 typedef enum MouseShapes { BEAM, HAND, ARROW } MouseShape;
 typedef enum { NONE, MENUBAR, WINDOW, ALL } WindowTitleIn;
+typedef enum { TILING, SCALED, MIRRORED } BackgroundImageLayout;
 
 #define MAX_CHILDREN 512
 #define BLANK_CHAR 0
-#define ATTRS_MASK_WITHOUT_WIDTH 0xFFC
+#define ATTRS_MASK_WITHOUT_WIDTH 0xFFFC
 #define WIDTH_MASK  3
 #define DECORATION_SHIFT  2
 #define DECORATION_MASK 3
@@ -68,6 +69,10 @@ typedef enum { NONE, MENUBAR, WINDOW, ALL } WindowTitleIn;
 #define REVERSE_SHIFT 6
 #define STRIKE_SHIFT 7
 #define DIM_SHIFT 8
+#define MARK_SHIFT 9
+#define ATTRS_MASK_WITHOUT_MARK 0xf9ff
+#define ATTRS_MASK_FOR_SGR (ATTRS_MASK_WITHOUT_MARK | ATTRS_MASK_WITHOUT_WIDTH)
+#define MARK_MASK 3
 #define COL_MASK 0xFFFFFFFF
 #define UTF8_ACCEPT 0
 #define UTF8_REJECT 1
@@ -141,6 +146,7 @@ typedef enum { NONE, MENUBAR, WINDOW, ALL } WindowTitleIn;
 #define END_ALLOW_UNUSED_RESULT _Pragma("GCC diagnostic pop")
 #endif
 
+
 typedef struct {
     uint32_t left, top, right, bottom;
 } Region;
@@ -231,6 +237,7 @@ typedef struct {
     DynamicColor dynamic_color_stack[10];
     size_t dynamic_color_stack_idx;
     DynamicColor configured, overridden;
+    color_type mark_foregrounds[MARK_MASK+1], mark_backgrounds[MARK_MASK+1];
 } ColorProfile;
 
 typedef struct {
@@ -246,17 +253,6 @@ typedef struct {FONTS_DATA_HEAD} *FONTS_DATA_HANDLE;
 
 #define clear_sprite_position(cell) (cell).sprite_x = 0; (cell).sprite_y = 0; (cell).sprite_z = 0;
 
-#define left_shift_line(line, at, num) { \
-    for(index_type __i__ = (at); __i__ < (line)->xnum - (num); __i__++) { \
-        COPY_CELL(line, __i__ + (num), line, __i__) \
-    } \
-    if ((((line)->gpu_cells[(at)].attrs) & WIDTH_MASK) != 1) { \
-        (line)->cpu_cells[(at)].ch = BLANK_CHAR; \
-        (line)->gpu_cells[(at)].attrs = BLANK_CHAR ? 1 : 0; \
-        clear_sprite_position((line)->gpu_cells[(at)]); \
-    }\
-}
-
 #define ensure_space_for(base, array, type, num, capacity, initial_cap, zero_mem) \
     if ((base)->capacity < num) { \
         size_t _newcap = MAX((size_t)initial_cap, MAX(2 * (base)->capacity, (size_t)num)); \
@@ -306,8 +302,7 @@ void set_mouse_cursor(MouseShape);
 void enter_event(void);
 void mouse_event(int, int, int);
 void focus_in_event(void);
-void scroll_event(double, double, int);
-void fake_scroll(int, bool);
+void scroll_event(double, double, int, int);
 void set_special_key_combo(int glfw_key, int mods, bool is_native);
 void on_key_input(GLFWkeyevent *ev);
 void request_window_attention(id_type, bool);
@@ -317,5 +312,10 @@ void play_canberra_sound(const char *which_sound, const char *event_id);
 SPRITE_MAP_HANDLE alloc_sprite_map(unsigned int, unsigned int);
 SPRITE_MAP_HANDLE free_sprite_map(SPRITE_MAP_HANDLE);
 
-static inline void safe_close(int fd) { while(close(fd) != 0 && errno == EINTR); }
+static inline void safe_close(int fd, const char* file UNUSED, const int line UNUSED) {
+#if 0
+    printf("Closing fd: %d from file: %s line: %d\n", fd, file, line);
+#endif
+    while(close(fd) != 0 && errno == EINTR);
+}
 void log_event(const char *format, ...) __attribute__((format(printf, 1, 2)));
