@@ -51,6 +51,12 @@ linebuf_mark_line_clean(LineBuf *self, index_type y) {
     self->line_attrs[y] &= ~TEXT_DIRTY_MASK;
 }
 
+void
+linebuf_mark_line_as_not_continued(LineBuf *self, index_type y) {
+    self->line_attrs[y] &= ~CONTINUED_MASK;
+}
+
+
 static PyObject*
 clear(LineBuf *self, PyObject *a UNUSED) {
 #define clear_doc "Clear all lines in this LineBuf"
@@ -421,14 +427,15 @@ as_ansi(LineBuf *self, PyObject *callback) {
 }
 
 static inline Line*
-get_line(LineBuf *self, index_type y) {
-    linebuf_init_line(self, y);
+get_line(void *x, int y) {
+    LineBuf *self = (LineBuf*)x;
+    linebuf_init_line(self, MAX(0, y));
     return self->line;
 }
 
 static PyObject*
 as_text(LineBuf *self, PyObject *args) {
-    as_text_generic(args, self, get_line, self->ynum, self->xnum);
+    return as_text_generic(args, self, get_line, self->ynum, self->xnum);
 }
 
 
@@ -529,8 +536,8 @@ linebuf_rewrap(LineBuf *self, LineBuf *other, index_type *num_content_lines_befo
     if (other->xnum == self->xnum && other->ynum == self->ynum) {
         memcpy(other->line_map, self->line_map, sizeof(index_type) * self->ynum);
         memcpy(other->line_attrs, self->line_attrs, sizeof(bool) * self->ynum);
-        memcpy(other->cpu_cell_buf, self->cpu_cell_buf, self->xnum * self->ynum * sizeof(CPUCell));
-        memcpy(other->gpu_cell_buf, self->gpu_cell_buf, self->xnum * self->ynum * sizeof(GPUCell));
+        memcpy(other->cpu_cell_buf, self->cpu_cell_buf, (size_t)self->xnum * self->ynum * sizeof(CPUCell));
+        memcpy(other->gpu_cell_buf, self->gpu_cell_buf, (size_t)self->xnum * self->ynum * sizeof(GPUCell));
         *num_content_lines_before = self->ynum; *num_content_lines_after = self->ynum;
         return;
     }
