@@ -33,7 +33,7 @@ clear_chars_in_line(CPUCell *cpu_cells, GPUCell *gpu_cells, index_type xnum, cha
 }
 
 static inline index_type
-xlimit_for_line(Line *line) {
+xlimit_for_line(const Line *line) {
     index_type xlimit = line->xnum;
     if (BLANK_CHAR == 0) {
         while (xlimit > 0 && (line->cpu_cells[xlimit - 1].ch) == BLANK_CHAR) xlimit--;
@@ -54,13 +54,28 @@ line_reset_cells(Line *line, index_type start, index_type num, GPUCell *gpu_cell
     memcpy(line->cpu_cells + start, cpu_cells + start, sizeof(CPUCell) * num);
 }
 
+static inline void
+left_shift_line(Line *line, index_type at, index_type num) {
+    for (index_type i = at; i < line->xnum - num; i++) {
+        COPY_CELL(line, i + num, line, i);
+    }
+    if (at < line->xnum && ((line->gpu_cells[at].attrs) & WIDTH_MASK) != 1) {
+        line->cpu_cells[at].ch = BLANK_CHAR;
+        line->gpu_cells[at].attrs = BLANK_CHAR ? 1 : 0;
+        clear_sprite_position(line->gpu_cells[at]);
+    }
+}
+
+typedef Line*(get_line_func)(void *, int);
 void line_clear_text(Line *self, unsigned int at, unsigned int num, char_type ch);
 void line_apply_cursor(Line *self, Cursor *cursor, unsigned int at, unsigned int num, bool clear_char);
+char_type line_get_char(Line *self, index_type at);
 void line_set_char(Line *, unsigned int , uint32_t , unsigned int , Cursor *, bool);
 void line_right_shift(Line *, unsigned int , unsigned int );
 void line_add_combining_char(Line *, uint32_t , unsigned int );
 index_type line_url_start_at(Line *self, index_type x);
-index_type line_url_end_at(Line *self, index_type x, bool, char_type);
+index_type line_url_end_at(Line *self, index_type x, bool, char_type, bool);
+bool line_startswith_url_chars(Line*);
 index_type line_as_ansi(Line *self, Py_UCS4 *buf, index_type buflen, bool*, const GPUCell**) __attribute__((nonnull));
 unsigned int line_length(Line *self);
 size_t cell_as_unicode(CPUCell *cell, bool include_cc, Py_UCS4 *buf, char_type);
@@ -81,6 +96,7 @@ void linebuf_set_attribute(LineBuf *, unsigned int , unsigned int );
 void linebuf_rewrap(LineBuf *self, LineBuf *other, index_type *, index_type *, HistoryBuf *, index_type *, index_type *);
 void linebuf_mark_line_dirty(LineBuf *self, index_type y);
 void linebuf_mark_line_clean(LineBuf *self, index_type y);
+void linebuf_mark_line_as_not_continued(LineBuf *self, index_type y);
 unsigned int linebuf_char_width_at(LineBuf *self, index_type x, index_type y);
 void linebuf_refresh_sprite_positions(LineBuf *self);
 void historybuf_add_line(HistoryBuf *self, const Line *line);
@@ -90,47 +106,6 @@ void historybuf_mark_line_clean(HistoryBuf *self, index_type y);
 void historybuf_mark_line_dirty(HistoryBuf *self, index_type y);
 void historybuf_refresh_sprite_positions(HistoryBuf *self);
 void historybuf_clear(HistoryBuf *self);
-
-
-#define as_text_generic(args, container, get_line, lines, columns) { \
-    PyObject *callback; \
-    int as_ansi = 0, insert_wrap_markers = 0; \
-    if (!PyArg_ParseTuple(args, "O|pp", &callback, &as_ansi, &insert_wrap_markers)) return NULL; \
-    PyObject *ret = NULL, *t = NULL; \
-    Py_UCS4 *buf = NULL; \
-    PyObject *nl = PyUnicode_FromString("\n"); \
-    PyObject *cr = PyUnicode_FromString("\r"); \
-    const GPUCell *prev_cell = NULL; \
-    if (nl == NULL || cr == NULL) goto end; \
-    if (as_ansi) { \
-        buf = malloc(sizeof(Py_UCS4) * columns * 100); \
-        if (buf == NULL) { PyErr_NoMemory(); goto end; } \
-    } \
-    for (index_type y = 0; y < lines; y++) { \
-        Line *line = get_line(container, y); \
-        if (!line->continued && y > 0) { \
-            ret = PyObject_CallFunctionObjArgs(callback, nl, NULL); \
-            if (ret == NULL) goto end; \
-            Py_CLEAR(ret); \
-        } \
-        if (as_ansi) { \
-            bool truncated; \
-            index_type num = line_as_ansi(line, buf, columns * 100 - 2, &truncated, &prev_cell); \
-            t = PyUnicode_FromKindAndData(PyUnicode_4BYTE_KIND, buf, num); \
-        } else { \
-            t = line_as_unicode(line); \
-        } \
-        if (t == NULL) goto end; \
-        ret = PyObject_CallFunctionObjArgs(callback, t, NULL); \
-        Py_DECREF(t); if (ret == NULL) goto end; Py_DECREF(ret); \
-        if (insert_wrap_markers) { \
-            ret = PyObject_CallFunctionObjArgs(callback, cr, NULL); \
-            if (ret == NULL) goto end; \
-            Py_CLEAR(ret); \
-        }\
-    } \
-end: \
-    Py_CLEAR(nl); Py_CLEAR(cr); free(buf); \
-    if (PyErr_Occurred()) return NULL; \
-    Py_RETURN_NONE; \
-}
+void mark_text_in_line(PyObject *marker, Line *line);
+bool line_has_mark(Line *, attrs_type mark);
+PyObject* as_text_generic(PyObject *args, void *container, get_line_func get_line, index_type lines, index_type columns);
