@@ -27,13 +27,26 @@ GlobalState global_state = {{0}};
 #define END_WITH_OS_WINDOW break; }}
 
 #define WITH_TAB(os_window_id, tab_id) \
-    for (size_t o = 0; o < global_state.num_os_windows; o++) { \
+    for (size_t o = 0, tab_found = 0; o < global_state.num_os_windows && !tab_found; o++) { \
         OSWindow *osw = global_state.os_windows + o; \
         if (osw->id == os_window_id) { \
             for (size_t t = 0; t < osw->num_tabs; t++) { \
                 if (osw->tabs[t].id == tab_id) { \
                     Tab *tab = osw->tabs + t;
-#define END_WITH_TAB break; }}}}
+#define END_WITH_TAB tab_found = 1; break; }}}}
+
+#define WITH_WINDOW(os_window_id, tab_id, window_id) \
+    for (size_t o = 0, window_found = 0; o < global_state.num_os_windows && !window_found; o++) { \
+        OSWindow *osw = global_state.os_windows + o; \
+        if (osw->id == os_window_id) { \
+            for (size_t t = 0; t < osw->num_tabs && !window_found; t++) { \
+                if (osw->tabs[t].id == tab_id) { \
+                    Tab *tab = osw->tabs + t; \
+                    for (size_t w = 0; w < tab->num_windows; w++) { \
+                        if (tab->windows[w].id == window_id) { \
+                            Window *window = tab->windows + w;
+#define END_WITH_WINDOW window_found = 1; break; }}}}}}
+
 
 #define WITH_OS_WINDOW_REFS \
     id_type cb_window_id = 0, focused_window_id = 0; \
@@ -71,6 +84,37 @@ os_window_for_kitty_window(id_type kitty_window_id) {
     return NULL;
 }
 
+static void
+send_bgimage_to_gpu(BackgroundImageLayout layout, BackgroundImage *bgimage) {
+    RepeatStrategy r;
+    switch (layout) {
+        case SCALED:
+            r = REPEAT_CLAMP; break;
+        case MIRRORED:
+            r = REPEAT_MIRROR; break;
+        case TILING:
+        default:
+            r = REPEAT_DEFAULT; break;
+    }
+    bgimage->texture_id = 0;
+    send_image_to_gpu(&bgimage->texture_id, bgimage->bitmap, bgimage->width,
+            bgimage->height, false, true, OPT(background_image_linear), r);
+    free(bgimage->bitmap); bgimage->bitmap = NULL;
+}
+
+static void
+free_bgimage(BackgroundImage **bgimage, bool release_texture) {
+    if (*bgimage && (*bgimage)->refcnt) {
+        (*bgimage)->refcnt--;
+        if ((*bgimage)->refcnt == 0) {
+            free((*bgimage)->bitmap); (*bgimage)->bitmap = NULL;
+            if (release_texture) free_texture(&(*bgimage)->texture_id);
+            free(*bgimage);
+        }
+    }
+    bgimage = NULL;
+}
+
 OSWindow*
 add_os_window() {
     WITH_OS_WINDOW_REFS
@@ -81,6 +125,24 @@ add_os_window() {
     ans->tab_bar_render_data.vao_idx = create_cell_vao();
     ans->gvao_idx = create_graphics_vao();
     ans->background_opacity = OPT(background_opacity);
+
+    bool wants_bg = OPT(background_image) && OPT(background_image)[0] != 0;
+    if (wants_bg) {
+        if (!global_state.bgimage) {
+            global_state.bgimage = calloc(1, sizeof(BackgroundImage));
+            if (!global_state.bgimage) fatal("Out of memory allocating the global bg image object");
+            global_state.bgimage->refcnt++;
+            size_t size;
+            if (png_path_to_bitmap(OPT(background_image), &global_state.bgimage->bitmap, &global_state.bgimage->width, &global_state.bgimage->height, &size)) {
+                send_bgimage_to_gpu(OPT(background_image_layout), global_state.bgimage);
+            }
+        }
+        if (global_state.bgimage->texture_id) {
+            ans->bgimage = global_state.bgimage;
+            ans->bgimage->refcnt++;
+        }
+    }
+
     ans->font_sz_in_pts = global_state.font_sz_in_pts;
     END_WITH_OS_WINDOW_REFS
     return ans;
@@ -107,10 +169,24 @@ create_gpu_resources_for_window(Window *w) {
 
 static inline void
 release_gpu_resources_for_window(Window *w) {
-    remove_vao(w->render_data.vao_idx); remove_vao(w->render_data.gvao_idx);
-    w->render_data.vao_idx = 0; w->render_data.gvao_idx = 0;
+    if (w->render_data.vao_idx > -1) remove_vao(w->render_data.vao_idx);
+    w->render_data.vao_idx = -1;
+    if (w->render_data.gvao_idx > -1) remove_vao(w->render_data.gvao_idx);
+    w->render_data.gvao_idx = -1;
 }
 
+static inline void
+initialize_window(Window *w, PyObject *title, bool init_gpu_resources) {
+    w->id = ++global_state.window_id_counter;
+    w->visible = true;
+    w->title = title;
+    Py_XINCREF(title);
+    if (init_gpu_resources) create_gpu_resources_for_window(w);
+    else {
+        w->render_data.vao_idx = -1;
+        w->render_data.gvao_idx = -1;
+    }
+}
 
 static inline id_type
 add_window(id_type os_window_id, id_type tab_id, PyObject *title) {
@@ -118,11 +194,7 @@ add_window(id_type os_window_id, id_type tab_id, PyObject *title) {
         ensure_space_for(tab, windows, Window, tab->num_windows + 1, capacity, 1, true);
         make_os_window_context_current(osw);
         zero_at_i(tab->windows, tab->num_windows);
-        tab->windows[tab->num_windows].id = ++global_state.window_id_counter;
-        tab->windows[tab->num_windows].visible = true;
-        tab->windows[tab->num_windows].title = title;
-        create_gpu_resources_for_window(&tab->windows[tab->num_windows]);
-        Py_INCREF(tab->windows[tab->num_windows].title);
+        initialize_window(tab->windows + tab->num_windows, title, true);
         return tab->windows[tab->num_windows++].id;
     END_WITH_TAB;
     return 0;
@@ -142,6 +214,27 @@ update_window_title(id_type os_window_id, id_type tab_id, id_type window_id, PyO
     END_WITH_TAB;
 }
 
+void
+set_os_window_title_from_window(Window *w, OSWindow *os_window) {
+    if (w->title && w->title != os_window->window_title) {
+        Py_XDECREF(os_window->window_title);
+        os_window->window_title = w->title;
+        Py_INCREF(os_window->window_title);
+        set_os_window_title(os_window, PyUnicode_AsUTF8(w->title));
+    }
+}
+
+void
+update_os_window_title(OSWindow *os_window) {
+    if (os_window->num_tabs) {
+        Tab *tab = os_window->tabs + os_window->active_tab;
+        if (tab->num_windows) {
+            Window *w = tab->windows + tab->active_window;
+            set_os_window_title_from_window(w, os_window);
+        }
+    }
+}
+
 static inline void
 destroy_window(Window *w) {
     Py_CLEAR(w->render_data.screen); Py_CLEAR(w->title);
@@ -150,7 +243,16 @@ destroy_window(Window *w) {
 
 static inline void
 remove_window_inner(Tab *tab, id_type id) {
+    id_type active_window_id = 0;
+    if (tab->active_window < tab->num_windows) active_window_id = tab->windows[tab->active_window].id;
     REMOVER(tab->windows, id, tab->num_windows, destroy_window, tab->capacity);
+    if (active_window_id) {
+        for (unsigned int w = 0; w < tab->num_windows; w++) {
+            if (tab->windows[w].id == active_window_id) {
+                tab->active_window = w; break;
+            }
+        }
+    }
 }
 
 static inline void
@@ -219,6 +321,7 @@ attach_window(id_type os_window_id, id_type tab_id, id_type id) {
                     w->render_data.screen->cell_size.height != osw->fonts_data->cell_height
                 ) resize_screen(osw, w->render_data.screen, true);
                 else screen_dirty_sprite_positions(w->render_data.screen);
+                w->render_data.screen->reload_all_gpu_data = true;
                 break;
             }
         }
@@ -235,8 +338,17 @@ destroy_tab(Tab *tab) {
 
 static inline void
 remove_tab_inner(OSWindow *os_window, id_type id) {
+    id_type active_tab_id = 0;
+    if (os_window->active_tab < os_window->num_tabs) active_tab_id = os_window->tabs[os_window->active_tab].id;
     make_os_window_context_current(os_window);
     REMOVER(os_window->tabs, id, os_window->num_tabs, destroy_tab, os_window->capacity);
+    if (active_tab_id) {
+        for (unsigned int i = 0; i < os_window->num_tabs; i++) {
+            if (os_window->tabs[i].id == active_tab_id) {
+                os_window->active_tab = i; break;
+            }
+        }
+    }
 }
 
 static inline void
@@ -254,9 +366,12 @@ destroy_os_window_item(OSWindow *w) {
     }
     Py_CLEAR(w->window_title); Py_CLEAR(w->tab_bar_render_data.screen);
     if (w->offscreen_texture_id) free_texture(&w->offscreen_texture_id);
+    if (w->offscreen_framebuffer) free_framebuffer(&w->offscreen_framebuffer);
     remove_vao(w->tab_bar_render_data.vao_idx);
     remove_vao(w->gvao_idx);
     free(w->tabs); w->tabs = NULL;
+    free_bgimage(&w->bgimage, true);
+    w->bgimage = NULL;
 }
 
 bool
@@ -285,11 +400,12 @@ set_active_tab(id_type os_window_id, unsigned int idx) {
 }
 
 static inline void
-set_active_window(id_type os_window_id, id_type tab_id, unsigned int idx) {
-    WITH_TAB(os_window_id, tab_id)
-        tab->active_window = idx;
+set_active_window(id_type os_window_id, id_type tab_id, id_type window_id) {
+    WITH_WINDOW(os_window_id, tab_id, window_id)
+        (void)window;
+        tab->active_window = w;
         osw->needs_render = true;
-    END_WITH_TAB;
+    END_WITH_WINDOW;
 }
 
 static inline void
@@ -301,15 +417,6 @@ swap_tabs(id_type os_window_id, unsigned int a, unsigned int b) {
     END_WITH_OS_WINDOW
 }
 
-static inline void
-swap_windows(id_type os_window_id, id_type tab_id, unsigned int a, unsigned int b) {
-    WITH_TAB(os_window_id, tab_id);
-    Window w = tab->windows[b];
-    tab->windows[b] = tab->windows[a];
-    tab->windows[a] = w;
-    END_WITH_TAB;
-}
-
 static void
 add_borders_rect(id_type os_window_id, id_type tab_id, uint32_t left, uint32_t top, uint32_t right, uint32_t bottom, uint32_t color) {
     WITH_TAB(os_window_id, tab_id)
@@ -347,6 +454,13 @@ os_window_regions(OSWindow *os_window, Region *central, Region *tab_bar) {
     }
 }
 
+void
+mark_os_window_for_close(OSWindow* w, CloseRequest cr) {
+    global_state.has_pending_closes = true;
+    w->close_request = cr;
+}
+
+
 
 // Python API {{{
 #define PYWRAP0(name) static PyObject* py##name(PYNOARG)
@@ -361,7 +475,9 @@ os_window_regions(OSWindow *os_window, Region *central, Region *tab_bar) {
 #define KI(name) PYWRAP1(name) { id_type a; unsigned int b; PA("KI", &a, &b); name(a, b); Py_RETURN_NONE; }
 #define KII(name) PYWRAP1(name) { id_type a; unsigned int b, c; PA("KII", &a, &b, &c); name(a, b, c); Py_RETURN_NONE; }
 #define KKI(name) PYWRAP1(name) { id_type a, b; unsigned int c; PA("KKI", &a, &b, &c); name(a, b, c); Py_RETURN_NONE; }
+#define KKK(name) PYWRAP1(name) { id_type a, b, c; PA("KKK", &a, &b, &c); name(a, b, c); Py_RETURN_NONE; }
 #define KKII(name) PYWRAP1(name) { id_type a, b; unsigned int c, d; PA("KKII", &a, &b, &c, &d); name(a, b, c, d); Py_RETURN_NONE; }
+#define KKKK(name) PYWRAP1(name) { id_type a, b, c, d; PA("KKKK", &a, &b, &c, &d); name(a, b, c, d); Py_RETURN_NONE; }
 #define KK5I(name) PYWRAP1(name) { id_type a, b; unsigned int c, d, e, f, g; PA("KKIIIII", &a, &b, &c, &d, &e, &f, &g); name(a, b, c, d, e, f, g); Py_RETURN_NONE; }
 #define BOOL_SET(name) PYWRAP1(set_##name) { global_state.name = PyObject_IsTrue(args); Py_RETURN_NONE; }
 
@@ -414,6 +530,30 @@ window_title_in(PyObject *title_in) {
     return ALL;
 }
 
+static BackgroundImageLayout
+bglayout(PyObject *layout_name) {
+    const char *name = PyUnicode_AsUTF8(layout_name);
+    switch(name[0]) {
+        case 't': return TILING;
+        case 'm': return MIRRORED;
+        case 's': return SCALED;
+        default: break;
+    }
+    return TILING;
+}
+
+static void
+background_image(PyObject *src) {
+    if (OPT(background_image)) free(OPT(background_image));
+    OPT(background_image) = NULL;
+    if (src == Py_None || !PyUnicode_Check(src)) return;
+    Py_ssize_t sz;
+    const char *s = PyUnicode_AsUTF8AndSize(src, &sz);
+    OPT(background_image) = calloc(sz + 1, 1);
+    if (OPT(background_image)) memcpy(OPT(background_image), s, sz);
+}
+
+
 static MouseShape
 pointer_shape(PyObject *shape_name) {
     const char *name = PyUnicode_AsUTF8(shape_name);
@@ -426,6 +566,34 @@ pointer_shape(PyObject *shape_name) {
     return BEAM;
 }
 
+static inline void
+free_url_prefixes(void) {
+    OPT(url_prefixes).num = 0;
+    OPT(url_prefixes).max_prefix_len = 0;
+    if (OPT(url_prefixes).values) {
+        free(OPT(url_prefixes.values));
+        OPT(url_prefixes).values = NULL;
+    }
+}
+
+static void
+set_url_prefixes(PyObject *up) {
+    free_url_prefixes();
+    OPT(url_prefixes).values = calloc(PyTuple_GET_SIZE(up), sizeof(UrlPrefix));
+    if (!OPT(url_prefixes).values) { PyErr_NoMemory(); return; }
+    OPT(url_prefixes).num = PyTuple_GET_SIZE(up);
+    for (size_t i = 0; i < OPT(url_prefixes).num; i++) {
+        PyObject *t = PyTuple_GET_ITEM(up, i);
+        if (!PyUnicode_Check(t)) { PyErr_SetString(PyExc_TypeError, "url_prefixes must be strings"); return; }
+        OPT(url_prefixes).values[i].len = MIN(arraysz(OPT(url_prefixes).values[i].string) - 1, (size_t)PyUnicode_GET_LENGTH(t));
+        int kind = PyUnicode_KIND(t);
+        OPT(url_prefixes).max_prefix_len = MAX(OPT(url_prefixes).max_prefix_len, OPT(url_prefixes).values[i].len);
+        for (size_t x = 0; x < OPT(url_prefixes).values[i].len; x++) {
+            OPT(url_prefixes).values[i].string[x] = PyUnicode_READ(kind, PyUnicode_DATA(t), x);
+        }
+    }
+}
+
 #define dict_iter(d) { \
     PyObject *key, *value; Py_ssize_t pos = 0; \
     while (PyDict_Next(d, &pos, &key, &value))
@@ -462,32 +630,36 @@ PYWRAP1(handle_for_window_id) {
 
 PYWRAP1(set_options) {
     PyObject *ret, *opts;
-    int is_wayland = 0, debug_gl = 0, debug_font_fallback = 0;
-    PA("O|ppp", &opts, &is_wayland, &debug_gl, &debug_font_fallback);
+    int is_wayland = 0, debug_rendering = 0, debug_font_fallback = 0;
+    PA("O|ppp", &opts, &is_wayland, &debug_rendering, &debug_font_fallback);
     global_state.is_wayland = is_wayland ? true : false;
 #ifdef __APPLE__
     global_state.has_render_frames = true;
 #endif
     if (global_state.is_wayland) global_state.has_render_frames = true;
-    global_state.debug_gl = debug_gl ? true : false;
+    global_state.debug_rendering = debug_rendering ? true : false;
     global_state.debug_font_fallback = debug_font_fallback ? true : false;
 #define GA(name) ret = PyObject_GetAttrString(opts, #name); if (ret == NULL) return NULL;
 #define SS(name, dest, convert) { GA(name); dest = convert(ret); Py_DECREF(ret); if (PyErr_Occurred()) return NULL; }
 #define S(name, convert) SS(name, OPT(name), convert)
     SS(kitty_mod, kitty_mod, PyLong_AsLong);
-    S(hide_window_decorations, PyObject_IsTrue);
+    S(hide_window_decorations, PyLong_AsUnsignedLong);
     S(visual_bell_duration, parse_s_double_to_monotonic_t);
     S(enable_audio_bell, PyObject_IsTrue);
     S(focus_follows_mouse, PyObject_IsTrue);
     S(cursor_blink_interval, parse_s_double_to_monotonic_t);
     S(cursor_stop_blinking_after, parse_s_double_to_monotonic_t);
     S(background_opacity, PyFloat_AsFloat);
+    S(background_image_layout, bglayout);
+    S(background_tint, PyFloat_AsFloat);
+    S(background_image_linear, PyObject_IsTrue);
     S(dim_opacity, PyFloat_AsFloat);
     S(dynamic_background_opacity, PyObject_IsTrue);
     S(inactive_text_alpha, PyFloat_AsFloat);
-    S(window_padding_width, PyFloat_AsFloat);
     S(scrollback_pager_history_size, PyLong_AsUnsignedLong);
     S(cursor_shape, PyLong_AsLong);
+    S(cursor_beam_thickness, PyFloat_AsFloat);
+    S(cursor_underline_thickness, PyFloat_AsFloat);
     S(url_style, PyLong_AsUnsignedLong);
     S(tab_bar_edge, PyLong_AsLong);
     S(mouse_hide_wait, parse_s_double_to_monotonic_t);
@@ -498,6 +670,12 @@ PYWRAP1(set_options) {
     S(terminal_select_modifiers, convert_mods);
     S(click_interval, parse_s_double_to_monotonic_t);
     S(resize_debounce_time, parse_s_double_to_monotonic_t);
+    S(mark1_foreground, color_as_int);
+    S(mark1_background, color_as_int);
+    S(mark2_foreground, color_as_int);
+    S(mark2_background, color_as_int);
+    S(mark3_foreground, color_as_int);
+    S(mark3_background, color_as_int);
     S(url_color, color_as_int);
     S(background, color_as_int);
     S(foreground, color_as_int);
@@ -520,7 +698,9 @@ PYWRAP1(set_options) {
     S(macos_thicken_font, PyFloat_AsFloat);
     S(tab_bar_min_tabs, PyLong_AsUnsignedLong);
     S(disable_ligatures, PyLong_AsLong);
+    S(force_ltr, PyObject_IsTrue);
     S(resize_draw_strategy, PyLong_AsLong);
+    S(resize_in_steps, PyObject_IsTrue);
     S(pointer_shape_when_grabbed, pointer_shape);
 
     GA(tab_bar_style);
@@ -528,6 +708,13 @@ PYWRAP1(set_options) {
     Py_CLEAR(ret);
     if (PyErr_Occurred()) return NULL;
 
+    PyObject *up = PyObject_GetAttrString(opts, "url_prefixes");
+    if (up == NULL) return NULL;
+    if (!PyTuple_Check(up)) { PyErr_SetString(PyExc_TypeError, "url_prefixes must be a tuple"); return NULL; }
+    set_url_prefixes(up);
+    Py_DECREF(up);
+    if (PyErr_Occurred()) return NULL;
+
     PyObject *chars = PyObject_GetAttrString(opts, "select_by_word_characters");
     if (chars == NULL) return NULL;
     for (size_t i = 0; i < MIN((size_t)PyUnicode_GET_LENGTH(chars), sizeof(OPT(select_by_word_characters))/sizeof(OPT(select_by_word_characters[0]))); i++) {
@@ -541,6 +728,8 @@ PYWRAP1(set_options) {
     GA(sequence_map); set_special_keys(ret);
     Py_DECREF(ret); if (PyErr_Occurred()) return NULL;
 
+    GA(background_image); background_image(ret); Py_CLEAR(ret);
+
 #define read_adjust(name) { \
     PyObject *al = PyObject_GetAttrString(opts, #name); \
     if (PyFloat_Check(al)) { \
@@ -624,17 +813,39 @@ end:
 }
 
 
+PYWRAP1(os_window_has_background_image) {
+    id_type os_window_id;
+    PA("K", &os_window_id);
+    WITH_OS_WINDOW(os_window_id)
+        if (os_window->bgimage && os_window->bgimage->texture_id > 0) { Py_RETURN_TRUE; }
+    END_WITH_OS_WINDOW
+    Py_RETURN_FALSE;
+}
+
 PYWRAP1(mark_os_window_for_close) {
     id_type os_window_id;
-    int yes = 1;
-    PA("K|p", &os_window_id, &yes);
+    CloseRequest cr = IMPERATIVE_CLOSE_REQUESTED;
+    PA("K|i", &os_window_id, &cr);
     WITH_OS_WINDOW(os_window_id)
-        mark_os_window_for_close(os_window, yes ? true : false);
+        mark_os_window_for_close(os_window, cr);
         Py_RETURN_TRUE;
     END_WITH_OS_WINDOW
     Py_RETURN_FALSE;
 }
 
+PYWRAP1(set_application_quit_request) {
+    CloseRequest cr = IMPERATIVE_CLOSE_REQUESTED;
+    PA("|i", &cr);
+    global_state.quit_request = cr;
+    global_state.has_pending_closes = true;
+    request_tick_callback();
+    Py_RETURN_NONE;
+}
+
+PYWRAP0(current_application_quit_request) {
+    return Py_BuildValue("i", global_state.quit_request);
+}
+
 PYWRAP1(focus_os_window) {
     id_type os_window_id;
     int also_raise = 1;
@@ -685,34 +896,32 @@ PYWRAP1(background_opacity_of) {
     Py_RETURN_NONE;
 }
 
-static inline bool
-fix_window_idx(Tab *tab, id_type window_id, unsigned int *window_idx) {
-    for (id_type fix = 0; fix < tab->num_windows; fix++) {
-        if (tab->windows[fix].id == window_id) { *window_idx = fix; return true; }
-    }
-    return false;
+PYWRAP1(set_window_padding) {
+    id_type os_window_id, tab_id, window_id;
+    unsigned int left, top, right, bottom;
+    PA("KKKIIII", &os_window_id, &tab_id, &window_id, &left, &top, &right, &bottom);
+    WITH_WINDOW(os_window_id, tab_id, window_id);
+        window->padding.left = left; window->padding.top = top; window->padding.right = right; window->padding.bottom = bottom;
+    END_WITH_WINDOW;
+    Py_RETURN_NONE;
 }
 
 PYWRAP1(set_window_render_data) {
 #define A(name) &(d.name)
 #define B(name) &(g.name)
     id_type os_window_id, tab_id, window_id;
-    unsigned int window_idx;
     ScreenRenderData d = {0};
     WindowGeometry g = {0};
-    PA("KKKIffffOIIII", &os_window_id, &tab_id, &window_id, &window_idx, A(xstart), A(ystart), A(dx), A(dy), A(screen), B(left), B(top), B(right), B(bottom));
+    PA("KKKffffOIIII", &os_window_id, &tab_id, &window_id, A(xstart), A(ystart), A(dx), A(dy), A(screen), B(left), B(top), B(right), B(bottom));
 
-    WITH_TAB(os_window_id, tab_id);
-        if (tab->windows[window_idx].id != window_id) {
-            if (!fix_window_idx(tab, window_id, &window_idx)) Py_RETURN_NONE;
-        }
-        Py_CLEAR(tab->windows[window_idx].render_data.screen);
-        d.vao_idx = tab->windows[window_idx].render_data.vao_idx;
-        d.gvao_idx = tab->windows[window_idx].render_data.gvao_idx;
-        tab->windows[window_idx].render_data = d;
-        tab->windows[window_idx].geometry = g;
-        Py_INCREF(tab->windows[window_idx].render_data.screen);
-    END_WITH_TAB;
+    WITH_WINDOW(os_window_id, tab_id, window_id);
+        Py_CLEAR(window->render_data.screen);
+        d.vao_idx = window->render_data.vao_idx;
+        d.gvao_idx = window->render_data.gvao_idx;
+        window->render_data = d;
+        window->geometry = g;
+        Py_INCREF(window->render_data.screen);
+    END_WITH_WINDOW;
     Py_RETURN_NONE;
 #undef A
 #undef B
@@ -720,18 +929,25 @@ PYWRAP1(set_window_render_data) {
 
 PYWRAP1(update_window_visibility) {
     id_type os_window_id, tab_id, window_id;
-    unsigned int window_idx;
     int visible;
-    PA("KKKIp", &os_window_id, &tab_id, &window_id, &window_idx, &visible);
-    WITH_TAB(os_window_id, tab_id);
-        if (tab->windows[window_idx].id != window_id) {
-            if (!fix_window_idx(tab, window_id, &window_idx)) Py_RETURN_NONE;
-        }
-        tab->windows[window_idx].visible = visible & 1;
-    END_WITH_TAB;
+    PA("KKKp", &os_window_id, &tab_id, &window_id, &visible);
+    WITH_WINDOW(os_window_id, tab_id, window_id);
+        window->visible = visible & 1;
+    END_WITH_WINDOW;
     Py_RETURN_NONE;
 }
 
+
+PYWRAP1(sync_os_window_title) {
+    id_type os_window_id;
+    PA("K", &os_window_id);
+    WITH_OS_WINDOW(os_window_id)
+        update_os_window_title(os_window);
+    END_WITH_OS_WINDOW
+    Py_RETURN_NONE;
+}
+
+
 static inline double
 dpi_for_os_window_id(id_type os_window_id) {
     double dpi = 0;
@@ -780,6 +996,7 @@ PYWRAP1(os_window_font_size) {
                     resize_screen(os_window, w->render_data.screen, true);
                 }
             }
+            if (OPT(resize_in_steps)) os_window_update_size_increments(os_window);
         }
         return Py_BuildValue("d", os_window->font_sz_in_pts);
     END_WITH_OS_WINDOW
@@ -793,6 +1010,14 @@ PYWRAP1(set_boss) {
     Py_RETURN_NONE;
 }
 
+PYWRAP0(get_boss) {
+    if (global_state.boss) {
+        Py_INCREF(global_state.boss);
+        return global_state.boss;
+    }
+    Py_RETURN_NONE;
+}
+
 PYWRAP1(patch_global_colors) {
     PyObject *spec;
     int configured;
@@ -806,17 +1031,85 @@ PYWRAP1(patch_global_colors) {
     P(active_border_color); P(inactive_border_color); P(bell_border_color);
     if (configured) {
         P(background); P(url_color);
+        P(mark1_background); P(mark1_foreground); P(mark2_background); P(mark2_foreground);
+        P(mark3_background); P(mark3_foreground);
     }
     if (PyErr_Occurred()) return NULL;
     Py_RETURN_NONE;
 }
 
+static PyObject*
+pyset_background_image(PyObject *self UNUSED, PyObject *args) {
+    const char *path;
+    PyObject *layout_name = NULL;
+    PyObject *os_window_ids;
+    int configured = 0;
+    PA("zO!|pU", &path, &PyTuple_Type, &os_window_ids, &configured, &layout_name);
+    size_t size;
+    BackgroundImageLayout layout = layout_name ? bglayout(layout_name) : OPT(background_image_layout);
+    BackgroundImage *bgimage = NULL;
+    if (path) {
+        bgimage = calloc(1, sizeof(BackgroundImage));
+        if (!bgimage) return PyErr_NoMemory();
+        if (!png_path_to_bitmap(path, &bgimage->bitmap, &bgimage->width, &bgimage->height, &size)) {
+            PyErr_Format(PyExc_ValueError, "Failed to load image from: %s", path);
+            free(bgimage);
+            return NULL;
+        }
+        send_bgimage_to_gpu(layout, bgimage);
+        bgimage->refcnt++;
+    }
+    if (configured) {
+        free_bgimage(&global_state.bgimage, true);
+        global_state.bgimage = bgimage;
+        if (bgimage) bgimage->refcnt++;
+        OPT(background_image_layout) = layout;
+    }
+    for (Py_ssize_t i = 0; i < PyTuple_GET_SIZE(os_window_ids); i++) {
+        id_type os_window_id = PyLong_AsUnsignedLongLong(PyTuple_GET_ITEM(os_window_ids, i));
+        WITH_OS_WINDOW(os_window_id)
+            make_os_window_context_current(os_window);
+            free_bgimage(&os_window->bgimage, true);
+            os_window->bgimage = bgimage;
+            os_window->render_calls = 0;
+            if (bgimage) bgimage->refcnt++;
+        END_WITH_OS_WINDOW
+    }
+    if (bgimage) free_bgimage(&bgimage, true);
+    Py_RETURN_NONE;
+}
+
 PYWRAP0(destroy_global_data) {
     Py_CLEAR(global_state.boss);
     free(global_state.os_windows); global_state.os_windows = NULL;
     Py_RETURN_NONE;
 }
 
+static void
+destroy_mock_window(PyObject *capsule) {
+    Window *w = PyCapsule_GetPointer(capsule, "Window");
+    if (w) {
+        destroy_window(w);
+        PyMem_Free(w);
+    }
+}
+
+static PyObject*
+pycreate_mock_window(PyObject *self UNUSED, PyObject *args) {
+    Screen *screen;
+    PyObject *title = NULL;
+    if (!PyArg_ParseTuple(args, "O|U", &screen, &title)) return NULL;
+    Window *w = PyMem_Calloc(sizeof(Window), 1);
+    if (!w) return NULL;
+    Py_INCREF(screen);
+    PyObject *ans = PyCapsule_New(w, "Window", destroy_mock_window);
+    if (ans != NULL) {
+        initialize_window(w, title, false);
+        w->render_data.screen = screen;
+    }
+    return ans;
+}
+
 THREE_ID_OBJ(update_window_title)
 THREE_ID(remove_window)
 THREE_ID(detach_window)
@@ -827,9 +1120,8 @@ PYWRAP1(add_window) { PyObject *title; id_type a, b; PA("KKO", &a, &b, &title);
 PYWRAP0(current_os_window) { OSWindow *w = current_os_window(); if (!w) Py_RETURN_NONE; return PyLong_FromUnsignedLongLong(w->id); }
 TWO_ID(remove_tab)
 KI(set_active_tab)
-KKI(set_active_window)
+KKK(set_active_window)
 KII(swap_tabs)
-KKII(swap_windows)
 KK5I(add_borders_rect)
 
 #define M(name, arg_type) {#name, (PyCFunction)name, arg_type, NULL}
@@ -853,23 +1145,30 @@ static PyMethodDef module_methods[] = {
     MW(set_active_tab, METH_VARARGS),
     MW(set_active_window, METH_VARARGS),
     MW(swap_tabs, METH_VARARGS),
-    MW(swap_windows, METH_VARARGS),
     MW(add_borders_rect, METH_VARARGS),
     MW(set_tab_bar_render_data, METH_VARARGS),
     MW(set_window_render_data, METH_VARARGS),
+    MW(set_window_padding, METH_VARARGS),
     MW(viewport_for_window, METH_VARARGS),
     MW(cell_size_for_window, METH_VARARGS),
+    MW(os_window_has_background_image, METH_VARARGS),
     MW(mark_os_window_for_close, METH_VARARGS),
+    MW(set_application_quit_request, METH_VARARGS),
+    MW(current_application_quit_request, METH_NOARGS),
     MW(set_titlebar_color, METH_VARARGS),
     MW(focus_os_window, METH_VARARGS),
     MW(mark_tab_bar_dirty, METH_O),
     MW(change_background_opacity, METH_VARARGS),
     MW(background_opacity_of, METH_O),
     MW(update_window_visibility, METH_VARARGS),
+    MW(sync_os_window_title, METH_VARARGS),
     MW(global_font_size, METH_VARARGS),
+    MW(set_background_image, METH_VARARGS),
     MW(os_window_font_size, METH_VARARGS),
     MW(set_boss, METH_O),
+    MW(get_boss, METH_NOARGS),
     MW(patch_global_colors, METH_VARARGS),
+    MW(create_mock_window, METH_VARARGS),
     MW(destroy_global_data, METH_NOARGS),
 
     {NULL, NULL, 0, NULL}        /* Sentinel */
@@ -882,6 +1181,14 @@ finalize(void) {
     }
     if (detached_windows.windows) free(detached_windows.windows);
     detached_windows.capacity = 0;
+    if (OPT(background_image)) free(OPT(background_image));
+    // we leak the texture here since it is not guaranteed
+    // that freeing the texture will work during shutdown and
+    // the GPU driver should take care of it when the OpenGL context is
+    // destroyed.
+    free_bgimage(&global_state.bgimage, false);
+    global_state.bgimage = NULL;
+    free_url_prefixes();
 }
 
 bool
@@ -897,6 +1204,9 @@ init_state(PyObject *module) {
     if (PyStructSequence_InitType2(&RegionType, &region_desc) != 0) return false;
     Py_INCREF((PyObject *) &RegionType);
     PyModule_AddObject(module, "Region", (PyObject *) &RegionType);
+    PyModule_AddIntConstant(module, "IMPERATIVE_CLOSE_REQUESTED", IMPERATIVE_CLOSE_REQUESTED);
+    PyModule_AddIntConstant(module, "NO_CLOSE_REQUESTED", NO_CLOSE_REQUESTED);
+    PyModule_AddIntConstant(module, "CLOSE_BEING_CONFIRMED", CLOSE_BEING_CONFIRMED);
     if (Py_AtExit(finalize) != 0) {
         PyErr_SetString(PyExc_RuntimeError, "Failed to register the state at exit handler");
         return false;
