@@ -232,12 +232,17 @@ class TestDataTypes(BaseTest):
         self.assertEqualAttributes(l3.cursor_from(0), q)
 
     def test_url_at(self):
+        self.set_options()
+
         def create(t):
             lb = create.lb = LineBuf(1, len(t))
             lf = lb.line(0)
             lf.set_text(t, 0, len(t), C())
             return lf
 
+        l0 = create('file:///etc/test')
+        self.ae(l0.url_start_at(0), 0)
+
         for trail in '.,]>)\\':
             lx = create("http://xyz.com" + trail)
             self.ae(lx.url_end_at(0), len(lx) - 2)
@@ -276,6 +281,14 @@ class TestDataTypes(BaseTest):
         l4 = create(' xxxxxtekljhgdkjgd')
         self.ae(l4.url_end_at(0), 0)
 
+        for trail in '/-&':
+            l4 = create('http://a.b?q=1' + trail)
+            self.ae(l4.url_end_at(1), len(l4) - 1)
+
+        l4 = create('http://a.b.')
+        self.ae(l4.url_end_at(0), len(l4) - 2)
+        self.ae(l4.url_end_at(0, 0, True), len(l4) - 1)
+
     def rewrap(self, lb, lb2):
         hb = HistoryBuf(lb2.ynum, lb2.xnum)
         cy = lb.rewrap(lb2, hb)
@@ -346,6 +359,10 @@ class TestDataTypes(BaseTest):
         self.ae(wcswidth('\033a\033[2mb'), 2)
         self.ae(wcswidth('\u25b6\ufe0f'), 2)
         self.ae(wcswidth('\U0001f610\ufe0e'), 1)
+        self.ae(wcswidth('\U0001f1e6a'), 3)
+        self.ae(wcswidth('\U0001F1E6a\U0001F1E8a'), 6)
+        self.ae(wcswidth('\U0001F1E6\U0001F1E8a'), 3)
+        self.ae(wcswidth('\U0001F1E6\U0001F1E8\U0001F1E6'), 4)
         # Regional indicator symbols (unicode flags) are defined as having
         # Emoji_Presentation so must have width 2
         self.ae(tuple(map(w, '\U0001f1ee\U0001f1f3')), (2, 2))
